#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <err.h>
#include <time.h>	/* clock_gettime */
#include <sys/time.h>	/* timespecsub */

#include <regex.h>

int
match(const regex_t *re, const char *input, size_t sz)
{
	static char buf[128];
	regmatch_t pmatch;
	int x;

	assert (input[sz] == '\0');
	x = regexec(re, input, 1, &pmatch, 0);
	if (x == REG_NOMATCH)
		return -1;
	else if (x != 0) {
		regerror(x, re, buf, sizeof buf);
		errx(100, "regexec: %s", buf);
	}

	return (int)(pmatch.rm_eo - pmatch.rm_so);
}

int
benchmark(regex_t *re)
{
	static const char *inputs[] = {
	    "test@example.com",
	    "hans.wurst+kaese@gmail.com",
	    "!#$%&@valid.email",
	    "...@invalid",
	    "total@falsch.",
	    "auch@.falsch",
	    "genauso.@falsch.ey",
	    "bis@hier.richtig ABER",
	    "mett@wurst",
	    "hallo@",
	    "@dingens",
	    "@",
	    "",
	    NULL };
	
	static const long N = 10000;		/* runs per input */

	struct timespec t, t0, td;
	const char **p;
	int i, n;
	size_t sz;

	/* warmup */
	sz = strlen(inputs[0]);
	for (i = 0; i < N; i++)
		match(re, inputs[0], sz);

	for (p = inputs; *p != NULL; p++) {
		printf("%30.30s  ", *p);
		fflush(stdout);

		sz = strlen(*p);
		clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &t0);
		for (i = 0; i < N; i++)
			n = match(re, *p, sz);
		clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &t);
		timespecsub(&t, &t0, &td);

		if (n == sz)
			printf("match     ");
		else if (n < 0)
			printf("no match  ");
		else
			printf("partial   ");

		printf("%9.0f ns\n", (td.tv_sec * 1e9 + td.tv_nsec) / N);
	}

	return 0;
}

#include <getopt.h>

/* email addresses. cf. rfc 5322, section 3.4 */
#define ATEXT	"[a-zA-Z0-9!#$%&'*+/=?^_`{|}~-]"
#define WORD	ATEXT "+"
#define DOTTED	WORD "(\\." WORD ")*"
#define ADDRESS	DOTTED "(@)" DOTTED

int
main(int argc, char **argv)
{
	int ch, x;
	char buf[128];
	regex_t re;

	if ((x = regcomp(&re, "^" ADDRESS, REG_EXTENDED)) != 0) {
		regerror(x, &re, buf, sizeof buf);
		fprintf(stderr, "regcomp: %s\n", buf);
		return 1;
	}

	while ((ch = getopt(argc, argv, "b")) != -1) {
		switch (ch) {
		case 'b':
			return benchmark(&re);
		default:
			printf("usage: nfa [-b] string ...\n");
			return 1;
		}
	}

	for (int i = 1; i < argc; i++) {
		char *inp;
		size_t sz;
		int n;

		inp = argv[i];
		sz = strlen(inp);
		n = match(&re, inp, sz);
		if (n == -1)
			printf("no ");
		else if (n < sz)
			printf("partial ");
		printf("match: %.*s\n", n, inp);
	}
	
	return 0;
}
